/*
    Copyright (C) 2001-2003 KSVG Team
    This file is part of the KDE project

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#include <math.h>

#include <kdebug.h>
#include <klocale.h>

#include "SVGRectImpl.h"
#include "SVGPaintImpl.h"
#include "SVGPointImpl.h"
#include "SVGAngleImpl.h"
#include "SVGDocumentImpl.h"
#include "SVGSVGElementImpl.h"
#include "SVGPathSegArcImpl.h"
#include "SVGPathSegListImpl.h"
#include "SVGPathElementImpl.h"
#include "SVGPathSegLinetoImpl.h"
#include "SVGPathSegMovetoImpl.h"
#include "SVGAnimatedNumberImpl.h"
#include "SVGPathSegClosePathImpl.h"
#include "SVGPathSegCurvetoCubicImpl.h"
#include "SVGPathSegLinetoVerticalImpl.h"
#include "SVGPathSegLinetoHorizontalImpl.h"
#include "SVGPathSegCurvetoQuadraticImpl.h"
#include "SVGPathSegCurvetoCubicSmoothImpl.h"
#include "SVGPathSegCurvetoQuadraticSmoothImpl.h"

#include "SVGPaint.h"

#include "CanvasItem.h"
#include "KSVGCanvas.h"
#include "BezierPath.h"
#include "Point.h"

using namespace KSVG;

#include "SVGPathElementImpl.lut.h"
#include "ksvg_scriptinterpreter.h"
#include "ksvg_bridge.h"
#include "ksvg_ecma.h"

SVGPathElementImpl::SVGPathElementImpl(DOM::ElementImpl *impl) : SVGShapeImpl(impl), SVGTestsImpl(), SVGLangSpaceImpl(), SVGExternalResourcesRequiredImpl(), SVGStylableImpl(this), SVGTransformableImpl(), SVGAnimatedPathDataImpl(), SVGPathParser()
{
	KSVG_EMPTY_FLAGS

	m_pathLength = new SVGAnimatedNumberImpl();
	m_pathLength->ref();

	m_pathLength->setBaseVal(0);
}

SVGPathElementImpl::~SVGPathElementImpl()
{
	pathSegList()->clear();

	if(m_pathLength)
		m_pathLength->deref();
}

SVGAnimatedNumberImpl *SVGPathElementImpl::pathLength() const
{
	return m_pathLength;
}

double SVGPathElementImpl::getTotalLength()
{
	T2P::BezierPath *path = ownerDoc()->canvas()->toBezierPath(m_item);
	if(path)
		return path->length();

	return 0;
}

SVGPointImpl *SVGPathElementImpl::getPointAtLength(double distance)
{
	SVGPointImpl *ret = SVGSVGElementImpl::createSVGPoint();
	double totalDistance = getTotalLength();
	T2P::BezierPath *path = ownerDoc()->canvas()->toBezierPath(m_item);
	if(path)
	{
		T2P::Point p;
		path->pointTangentNormalAt(distance / totalDistance, &p);
		ret->setX(p.x());
		ret->setY(p.y());
	}

	return ret;
}

unsigned long SVGPathElementImpl::getPathSegAtLength(double)
{
	return 0;
}

SVGPathSegClosePathImpl *SVGPathElementImpl::createSVGPathSegClosePath()
{
	SVGPathSegClosePathImpl *temp = new SVGPathSegClosePathImpl();
	temp->ref();

	return temp;
}

SVGPathSegMovetoAbsImpl *SVGPathElementImpl::createSVGPathSegMovetoAbs(double x, double y)
{
	SVGPathSegMovetoAbsImpl *temp = new SVGPathSegMovetoAbsImpl();
	temp->ref();

	temp->setX(x);
	temp->setY(y);
	return temp;
}

SVGPathSegMovetoRelImpl *SVGPathElementImpl::createSVGPathSegMovetoRel(double x, double y)
{
	SVGPathSegMovetoRelImpl *temp = new SVGPathSegMovetoRelImpl();
	temp->ref();

	temp->setX(x);
	temp->setY(y);
	return temp;
}

SVGPathSegLinetoAbsImpl *SVGPathElementImpl::createSVGPathSegLinetoAbs(double x, double y)
{
	SVGPathSegLinetoAbsImpl *temp = new SVGPathSegLinetoAbsImpl();
	temp->ref();

	temp->setX(x);
	temp->setY(y);
	return temp;
}

SVGPathSegLinetoRelImpl *SVGPathElementImpl::createSVGPathSegLinetoRel(double x, double y)
{
	SVGPathSegLinetoRelImpl *temp = new SVGPathSegLinetoRelImpl();
	temp->ref();

	temp->setX(x);
	temp->setY(y);
	return temp;
}

SVGPathSegCurvetoCubicAbsImpl *SVGPathElementImpl::createSVGPathSegCurvetoCubicAbs(double x, double y, double x1, double y1, double x2, double y2)
{
	SVGPathSegCurvetoCubicAbsImpl *temp = new SVGPathSegCurvetoCubicAbsImpl();
	temp->ref();

	temp->setX(x);
	temp->setY(y);
	temp->setX1(x1);
	temp->setY1(y1);
	temp->setX2(x2);
	temp->setY2(y2);
	return temp;
}

SVGPathSegCurvetoCubicRelImpl *SVGPathElementImpl::createSVGPathSegCurvetoCubicRel(double x, double y, double x1, double y1, double x2, double y2)
{
	SVGPathSegCurvetoCubicRelImpl *temp = new SVGPathSegCurvetoCubicRelImpl();
	temp->ref();

	temp->setX(x);
	temp->setY(y);
	temp->setX1(x1);
	temp->setY1(y1);
	temp->setX2(x2);
	temp->setY2(y2);
	return temp;
}

SVGPathSegCurvetoQuadraticAbsImpl *SVGPathElementImpl::createSVGPathSegCurvetoQuadraticAbs(double x, double y, double x1, double y1)
{
	SVGPathSegCurvetoQuadraticAbsImpl *temp = new SVGPathSegCurvetoQuadraticAbsImpl();
	temp->ref();

	temp->setX(x);
	temp->setY(y);
	temp->setX1(x1);
	temp->setY1(y1);
	return temp;
}

SVGPathSegCurvetoQuadraticRelImpl *SVGPathElementImpl::createSVGPathSegCurvetoQuadraticRel(double x, double y, double x1, double y1)
{
	SVGPathSegCurvetoQuadraticRelImpl *temp = new SVGPathSegCurvetoQuadraticRelImpl();
	temp->ref();

	temp->setX(x);
	temp->setY(y);
	temp->setX1(x1);
	temp->setY1(y1);
	return temp;
}

SVGPathSegArcAbsImpl *SVGPathElementImpl::createSVGPathSegArcAbs(double x, double y, double r1, double r2, double angle, bool largeArcFlag, bool sweepFlag)
{
	SVGPathSegArcAbsImpl *temp = new SVGPathSegArcAbsImpl();
	temp->ref();

	temp->setX(x);
	temp->setY(y);
	temp->setR1(r1);
	temp->setR2(r2);
	temp->setAngle(angle);
	temp->setLargeArcFlag(largeArcFlag);
	temp->setSweepFlag(sweepFlag);
	return temp;
}

SVGPathSegArcRelImpl *SVGPathElementImpl::createSVGPathSegArcRel(double x, double y, double r1, double r2, double angle, bool largeArcFlag, bool sweepFlag)
{
	SVGPathSegArcRelImpl *temp = new SVGPathSegArcRelImpl();
	temp->ref();

	temp->setX(x);
	temp->setY(y);
	temp->setR1(r1);
	temp->setR2(r2);
	temp->setAngle(angle);
	temp->setLargeArcFlag(largeArcFlag);
	temp->setSweepFlag(sweepFlag);
	return temp;
}

SVGPathSegLinetoHorizontalAbsImpl *SVGPathElementImpl::createSVGPathSegLinetoHorizontalAbs(double x)
{
	SVGPathSegLinetoHorizontalAbsImpl *temp = new SVGPathSegLinetoHorizontalAbsImpl();
	temp->ref();

	temp->setX(x);
	return temp;
}

SVGPathSegLinetoHorizontalRelImpl *SVGPathElementImpl::createSVGPathSegLinetoHorizontalRel(double x)
{
	SVGPathSegLinetoHorizontalRelImpl *temp = new SVGPathSegLinetoHorizontalRelImpl();
	temp->ref();

	temp->setX(x);
	return temp;
}

SVGPathSegLinetoVerticalAbsImpl *SVGPathElementImpl::createSVGPathSegLinetoVerticalAbs(double y)
{
	SVGPathSegLinetoVerticalAbsImpl *temp = new SVGPathSegLinetoVerticalAbsImpl();
	temp->ref();

	temp->setY(y);
	return temp;
}

SVGPathSegLinetoVerticalRelImpl *SVGPathElementImpl::createSVGPathSegLinetoVerticalRel(double y)
{
	SVGPathSegLinetoVerticalRelImpl *temp = new SVGPathSegLinetoVerticalRelImpl();
	temp->ref();

	temp->setY(y);
	return temp;
}

SVGPathSegCurvetoCubicSmoothAbsImpl *SVGPathElementImpl::createSVGPathSegCurvetoCubicSmoothAbs(double x, double y, double x2, double y2)
{
	SVGPathSegCurvetoCubicSmoothAbsImpl *temp = new SVGPathSegCurvetoCubicSmoothAbsImpl();
	temp->ref();

	temp->setX(x);
	temp->setY(y);
	temp->setX2(x2);
	temp->setY2(y2);
	return temp;
}

SVGPathSegCurvetoCubicSmoothRelImpl *SVGPathElementImpl::createSVGPathSegCurvetoCubicSmoothRel(double x, double y, double x2, double y2)
{
	SVGPathSegCurvetoCubicSmoothRelImpl *temp = new SVGPathSegCurvetoCubicSmoothRelImpl();
	temp->ref();

	temp->setX(x);
	temp->setY(y);
	temp->setX2(x2);
	temp->setY2(y2);
	return temp;
}

SVGPathSegCurvetoQuadraticSmoothAbsImpl *SVGPathElementImpl::createSVGPathSegCurvetoQuadraticSmoothAbs(double x, double y)
{
	SVGPathSegCurvetoQuadraticSmoothAbsImpl *temp = new SVGPathSegCurvetoQuadraticSmoothAbsImpl();
	temp->ref();

	temp->setX(x);
	temp->setY(y);
	return temp;
}

SVGPathSegCurvetoQuadraticSmoothRelImpl *SVGPathElementImpl::createSVGPathSegCurvetoQuadraticSmoothRel(double x, double y)
{
	SVGPathSegCurvetoQuadraticSmoothRelImpl *temp = new SVGPathSegCurvetoQuadraticSmoothRelImpl();
	temp->ref();

	temp->setX(x);
	temp->setY(y);
	return temp;
}

void SVGPathElementImpl::svgMoveTo(double x1, double y1, bool, bool abs)
{
	if(abs)
		pathSegList()->appendItem(createSVGPathSegMovetoAbs(x1, y1));
	else
		pathSegList()->appendItem(createSVGPathSegMovetoRel(x1, y1));
}

void SVGPathElementImpl::svgLineTo(double x1, double y1, bool abs)
{
	if(abs)
		pathSegList()->appendItem(createSVGPathSegLinetoAbs(x1, y1));
	else
		pathSegList()->appendItem(createSVGPathSegLinetoRel(x1, y1));
}

void SVGPathElementImpl::svgLineToHorizontal(double x, bool abs)
{
	if(abs)
		pathSegList()->appendItem(createSVGPathSegLinetoHorizontalAbs(x));
	else
		pathSegList()->appendItem(createSVGPathSegLinetoHorizontalRel(x));
}

void SVGPathElementImpl::svgLineToVertical(double y, bool abs)
{
	if(abs)
		pathSegList()->appendItem(createSVGPathSegLinetoVerticalAbs(y));
	else
		pathSegList()->appendItem(createSVGPathSegLinetoVerticalRel(y));
}

void SVGPathElementImpl::svgCurveToCubic(double x1, double y1, double x2, double y2, double x, double y, bool abs)
{
	if(abs)
		pathSegList()->appendItem(createSVGPathSegCurvetoCubicAbs(x, y, x1, y1, x2, y2));
	else
		pathSegList()->appendItem(createSVGPathSegCurvetoCubicRel(x, y, x1, y1, x2, y2));
}

void SVGPathElementImpl::svgCurveToCubicSmooth(double x, double y, double x2, double y2, bool abs)
{
	if(abs)
		pathSegList()->appendItem(createSVGPathSegCurvetoCubicSmoothAbs(x2, y2, x, y));
	else
		pathSegList()->appendItem(createSVGPathSegCurvetoCubicSmoothRel(x2, y2, x, y));
}

void SVGPathElementImpl::svgCurveToQuadratic(double x, double y, double x1, double y1, bool abs)
{
	if(abs)
		pathSegList()->appendItem(createSVGPathSegCurvetoQuadraticAbs(x1, y1, x, y));
	else
		pathSegList()->appendItem(createSVGPathSegCurvetoQuadraticRel(x1, y1, x, y));
}

void SVGPathElementImpl::svgCurveToQuadraticSmooth(double x, double y, bool abs)
{
	if(abs)
		pathSegList()->appendItem(createSVGPathSegCurvetoQuadraticSmoothAbs(x, y));
	else
		pathSegList()->appendItem(createSVGPathSegCurvetoQuadraticSmoothRel(x, y));
}

void SVGPathElementImpl::svgArcTo(double x, double y, double r1, double r2, double angle, bool largeArcFlag, bool sweepFlag, bool abs)
{
	if(abs)
		pathSegList()->appendItem(createSVGPathSegArcAbs(x, y, r1, r2, angle, largeArcFlag, sweepFlag));
	else
		pathSegList()->appendItem(createSVGPathSegArcRel(x, y, r1, r2, angle, largeArcFlag, sweepFlag));
}

void SVGPathElementImpl::svgClosePath()
{
	pathSegList()->appendItem(createSVGPathSegClosePath());
}

// Ecma stuff

/*
@begin SVGPathElementImpl::s_hashTable 3
 d											SVGPathElementImpl::D										DontDelete|ReadOnly
 pathLength									SVGPathElementImpl::PathLength								DontDelete|ReadOnly
@end
@begin SVGPathElementImplProto::s_hashTable 23
 getTotalLength								SVGPathElementImpl::GetTotalLength							DontDelete|Function 0
 getPointAtLength							SVGPathElementImpl::GetPointAtLength						DontDelete|Function 1
 getPathSegAtLength							SVGPathElementImpl::GetPathSegAtLength						DontDelete|Function 1
 createSVGPathSegClosePath					SVGPathElementImpl::CreateSVGPathSegClosePath				DontDelete|Function 0
 createSVGPathSegMovetoAbs					SVGPathElementImpl::CreateSVGPathSegMovetoAbs				DontDelete|Function 2
 createSVGPathSegMovetoRel					SVGPathElementImpl::CreateSVGPathSegMovetoRel				DontDelete|Function 2
 createSVGPathSegLinetoAbs					SVGPathElementImpl::CreateSVGPathSegLinetoAbs				DontDelete|Function 2
 createSVGPathSegLinetoRel					SVGPathElementImpl::CreateSVGPathSegLinetoRel				DontDelete|Function 2
 createSVGPathSegArcAbs						SVGPathElementImpl::CreateSVGPathSegArcAbs					DontDelete|Function 7
 createSVGPathSegArcRel						SVGPathElementImpl::CreateSVGPathSegArcRel					DontDelete|Function 7
 createSVGPathSegCurvetoCubicAbs			SVGPathElementImpl::CreateSVGPathSegCurvetoCubicAbs			DontDelete|Function 6
 createSVGPathSegCurvetoCubicRel			SVGPathElementImpl::CreateSVGPathSegCurvetoCubicRel			DontDelete|Function 6
 createSVGPathSegCurvetoQuadraticAbs		SVGPathElementImpl::CreateSVGPathSegCurvetoQuadraticAbs		DontDelete|Function 4
 createSVGPathSegCurvetoQuadraticRel		SVGPathElementImpl::CreateSVGPathSegCurvetoQuadraticRel		DontDelete|Function 4
 createSVGPathSegLinetoHorizontalAbs		SVGPathElementImpl::CreateSVGPathSegLinetoHorizontalAbs		DontDelete|Function 1
 createSVGPathSegLinetoHorizontalRel		SVGPathElementImpl::CreateSVGPathSegLinetoHorizontalRel		DontDelete|Function 1
 createSVGPathSegLinetoVerticalAbs			SVGPathElementImpl::CreateSVGPathSegLinetoVerticalAbs		DontDelete|Function 1
 createSVGPathSegLinetoVerticalRel			SVGPathElementImpl::CreateSVGPathSegLinetoVerticalRel		DontDelete|Function 1
 createSVGPathSegCurvetoCubicAbs			SVGPathElementImpl::CreateSVGPathSegCurvetoCubicAbs			DontDelete|Function 4
 createSVGPathSegCurvetoCubicRel			SVGPathElementImpl::CreateSVGPathSegCurvetoCubicRel			DontDelete|Function 4
 createSVGPathSegCurvetoQuadraticAbs		SVGPathElementImpl::CreateSVGPathSegCurvetoQuadraticAbs		DontDelete|Function 2
 createSVGPathSegCurvetoQuadraticRel		SVGPathElementImpl::CreateSVGPathSegCurvetoQuadraticRel		DontDelete|Function 2
@end
*/

KSVG_IMPLEMENT_PROTOTYPE("SVGPathElementImpl", SVGPathElementImplProto, SVGPathElementImplProtoFunc)

Value SVGPathElementImpl::getValueProperty(ExecState *exec, int token) const
{
	//KSVG_CHECK_ATTRIBUTE
		
	switch(token)
	{
		case PathLength:
			return m_pathLength->cache(exec);
		case D:
//	if(!attributeMode)
		{
				QString d;
				unsigned int nrSegs = pathSegList()->numberOfItems();
				SVGPathSegImpl *curseg = 0;
				for(unsigned int i = 0; i < nrSegs; i++)
				{
					curseg = pathSegList()->getItem(i);
					if(curseg)
						d += curseg->toString() + " ";
				}
				return String(d);
		}
		default:
			kdWarning() << "Unhandled token in " << k_funcinfo << " : " << token << endl;
			return Undefined();
	}
}

void SVGPathElementImpl::putValueProperty(ExecState *exec, int token, const Value &value, int attr)
{
	// This class has just ReadOnly properties, only with the Internal flag set
	// it's allowed to modify those.
	if(!(attr & KJS::Internal))
		return;
	
	switch(token)
	{
		case D:
		{
			pathSegList()->clear();
			QString d = value.toString(exec).qstring();
			parseSVG(d, false);
			break;
		}
		default:
			kdWarning() << "Unhandled token in " << k_funcinfo << " : " << token << endl;
	}
}

Value SVGPathElementImplProtoFunc::call(ExecState *exec, Object &thisObj, const List &args)
{
	KSVG_CHECK_THIS(SVGPathElementImpl)

	switch(id)
	{
		case SVGPathElementImpl::GetTotalLength:
			return Number(obj->getTotalLength());
		case SVGPathElementImpl::GetPointAtLength:
			return obj->getPointAtLength(args[0].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::GetPathSegAtLength:
			return Number(obj->getPathSegAtLength(args[0].toNumber(exec)));
		case SVGPathElementImpl::CreateSVGPathSegClosePath:
			return obj->createSVGPathSegClosePath()->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegMovetoAbs:
			return obj->createSVGPathSegMovetoAbs(args[0].toNumber(exec), args[1].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegMovetoRel:
			return obj->createSVGPathSegMovetoRel(args[0].toNumber(exec), args[1].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegLinetoAbs:
			return obj->createSVGPathSegLinetoAbs(args[0].toNumber(exec), args[1].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegLinetoRel:
			return obj->createSVGPathSegLinetoRel(args[0].toNumber(exec), args[1].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegCurvetoCubicAbs:
			return obj->createSVGPathSegCurvetoCubicAbs(args[0].toNumber(exec), args[1].toNumber(exec), args[2].toNumber(exec), args[3].toNumber(exec), args[4].toNumber(exec), args[5].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegCurvetoCubicRel:
			return obj->createSVGPathSegCurvetoCubicRel(args[0].toNumber(exec), args[1].toNumber(exec), args[2].toNumber(exec), args[3].toNumber(exec), args[4].toNumber(exec), args[5].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegCurvetoQuadraticAbs:
			return obj->createSVGPathSegCurvetoQuadraticAbs(args[0].toNumber(exec), args[1].toNumber(exec), args[2].toNumber(exec), args[3].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegCurvetoQuadraticRel:
			return obj->createSVGPathSegCurvetoQuadraticRel(args[0].toNumber(exec), args[1].toNumber(exec), args[2].toNumber(exec), args[3].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegArcAbs:
			return obj->createSVGPathSegArcAbs(args[0].toNumber(exec), args[1].toNumber(exec), args[2].toNumber(exec), args[3].toNumber(exec), args[4].toNumber(exec), args[5].toBoolean(exec), args[6].toBoolean(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegArcRel:
			return obj->createSVGPathSegArcRel(args[0].toNumber(exec), args[1].toNumber(exec), args[2].toNumber(exec), args[3].toNumber(exec), args[4].toNumber(exec), args[5].toBoolean(exec), args[6].toBoolean(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegLinetoHorizontalAbs:
			return obj->createSVGPathSegLinetoHorizontalAbs(args[0].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegLinetoHorizontalRel:
			return obj->createSVGPathSegLinetoHorizontalRel(args[0].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegLinetoVerticalAbs:
			return obj->createSVGPathSegLinetoVerticalAbs(args[0].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegLinetoVerticalRel:
			return obj->createSVGPathSegLinetoVerticalRel(args[0].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegCurvetoCubicSmoothAbs:
			return obj->createSVGPathSegCurvetoCubicSmoothAbs(args[0].toNumber(exec), args[1].toNumber(exec), args[2].toNumber(exec), args[3].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegCurvetoCubicSmoothRel:
			return obj->createSVGPathSegCurvetoCubicSmoothRel(args[0].toNumber(exec), args[1].toNumber(exec), args[2].toNumber(exec), args[3].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegCurvetoQuadraticSmoothAbs:
			return obj->createSVGPathSegCurvetoQuadraticSmoothAbs(args[0].toNumber(exec), args[1].toNumber(exec))->cache(exec);
		case SVGPathElementImpl::CreateSVGPathSegCurvetoQuadraticSmoothRel:
			return obj->createSVGPathSegCurvetoQuadraticSmoothRel(args[0].toNumber(exec), args[1].toNumber(exec))->cache(exec);
		default:
			kdWarning() << "Unhandled function id in " << k_funcinfo << " : " << id << endl;
			break;
	}

	return Undefined();
}

SVGRectImpl *SVGPathElementImpl::getBBox()
{
	SVGRectImpl *ret = SVGSVGElementImpl::createSVGRect();

	if(m_item)
	{
		T2P::BezierPath *path = ownerDoc()->canvas()->toBezierPath(m_item);
		if(path)
		{
			T2P::Point topLeft;
			T2P::Point bottomRight;

			path->boundingBox(&topLeft, &bottomRight);

			ret->setX(topLeft.x());
			ret->setY(topLeft.y());
			ret->setWidth(bottomRight.x() - topLeft.x());
			ret->setHeight(bottomRight.y() - topLeft.y());
		}
	}
	return ret;
}

void SVGPathElementImpl::createItem(KSVGCanvas *c)
{
	if(!c)
		c = ownerDoc()->canvas();

	if(!m_item)
	{
		m_item = c->createPath(this);
		c->insert(m_item);
	}
}

double SVGPathElementImpl::computeInSlope(float curx, float cury, SVGPathSegImpl *cur)
{
	switch(cur->pathSegType())
	{
		case PATHSEG_LINETO_ABS:
			return SVGAngleImpl::todeg(atan2(static_cast<SVGPathSegLinetoRelImpl *>(cur)->y() - cury,
											 static_cast<SVGPathSegLinetoRelImpl *>(cur)->x() - curx));
		case PATHSEG_LINETO_REL:
			return SVGAngleImpl::todeg(atan2(static_cast<SVGPathSegLinetoRelImpl *>(cur)->y(),
											 static_cast<SVGPathSegLinetoRelImpl *>(cur)->x()));
		case PATHSEG_LINETO_VERTICAL_ABS:
			return SVGAngleImpl::todeg(atan2(static_cast<SVGPathSegLinetoVerticalAbsImpl *>(cur)->y() - cury, 0));
		case PATHSEG_LINETO_VERTICAL_REL:
			return SVGAngleImpl::todeg(atan2(static_cast<SVGPathSegLinetoVerticalAbsImpl *>(cur)->y(), 0));
		case PATHSEG_LINETO_HORIZONTAL_ABS:
			return SVGAngleImpl::todeg(atan2(0, static_cast<SVGPathSegLinetoHorizontalAbsImpl *>(cur)->x() - curx));
		case PATHSEG_LINETO_HORIZONTAL_REL:
			return SVGAngleImpl::todeg(atan2(0, static_cast<SVGPathSegLinetoHorizontalAbsImpl *>(cur)->x()));
		case PATHSEG_CURVETO_CUBIC_ABS:
			return SVGAngleImpl::todeg(atan2(static_cast<SVGPathSegCurvetoCubicAbsImpl *>(cur)->y() - static_cast<SVGPathSegCurvetoCubicAbsImpl *>(cur)->y2(),
											 static_cast<SVGPathSegCurvetoCubicAbsImpl *>(cur)->x() - static_cast<SVGPathSegCurvetoCubicAbsImpl *>(cur)->x2()));
		case PATHSEG_CURVETO_CUBIC_REL:
			return SVGAngleImpl::todeg(atan2(static_cast<SVGPathSegCurvetoCubicRelImpl *>(cur)->y(),
											 static_cast<SVGPathSegCurvetoCubicRelImpl *>(cur)->x()));
	}
	
	return 0;
}

double SVGPathElementImpl::computeOutSlope(float curx, float cury, SVGPathSegImpl *next)
{
	switch(next->pathSegType())
	{
		case PATHSEG_LINETO_ABS:
			return SVGAngleImpl::todeg(atan2(static_cast<SVGPathSegLinetoRelImpl *>(next)->y() - cury,
											 static_cast<SVGPathSegLinetoRelImpl *>(next)->x() - curx));
		case PATHSEG_LINETO_REL:
			return SVGAngleImpl::todeg(atan2(static_cast<SVGPathSegLinetoRelImpl *>(next)->y(),
											 static_cast<SVGPathSegLinetoRelImpl *>(next)->x()));
		case PATHSEG_LINETO_VERTICAL_ABS:
			return SVGAngleImpl::todeg(atan2(static_cast<SVGPathSegLinetoVerticalAbsImpl *>(next)->y() - cury, 0));
		case PATHSEG_LINETO_VERTICAL_REL:
			return SVGAngleImpl::todeg(atan2(static_cast<SVGPathSegLinetoVerticalAbsImpl *>(next)->y(), 0));
		case PATHSEG_LINETO_HORIZONTAL_ABS:
			return SVGAngleImpl::todeg(atan2(0, static_cast<SVGPathSegLinetoHorizontalAbsImpl *>(next)->x() - curx));
		case PATHSEG_LINETO_HORIZONTAL_REL:
			return SVGAngleImpl::todeg(atan2(0, static_cast<SVGPathSegLinetoHorizontalAbsImpl *>(next)->x()));
		case PATHSEG_CURVETO_CUBIC_ABS:
			return SVGAngleImpl::todeg(atan2(static_cast<SVGPathSegCurvetoCubicAbsImpl *>(next)->y1() - cury,
											 static_cast<SVGPathSegCurvetoCubicAbsImpl *>(next)->x1() - curx));
		case PATHSEG_CURVETO_CUBIC_REL:
			return SVGAngleImpl::todeg(atan2(static_cast<SVGPathSegCurvetoCubicRelImpl *>(next)->y1(),
											 static_cast<SVGPathSegCurvetoCubicRelImpl *>(next)->x1()));
	}
	
	return 0;
}

void SVGPathElementImpl::computePos(SVGPathSegImpl *pathseg, float &curx, float &cury)
{
	switch(pathseg->pathSegType())
	{
		case PATHSEG_MOVETO_ABS:
			curx = static_cast<SVGPathSegMovetoRelImpl *>(pathseg)->x();
			cury = static_cast<SVGPathSegMovetoRelImpl *>(pathseg)->y();
			break;
		case PATHSEG_MOVETO_REL:
			curx += static_cast<SVGPathSegMovetoRelImpl *>(pathseg)->x();
			cury += static_cast<SVGPathSegMovetoRelImpl *>(pathseg)->y();
			break;
		case PATHSEG_LINETO_ABS:
			curx = static_cast<SVGPathSegLinetoRelImpl *>(pathseg)->x();
			cury = static_cast<SVGPathSegLinetoRelImpl *>(pathseg)->y();
			break;
		case PATHSEG_LINETO_REL:
			curx += static_cast<SVGPathSegLinetoRelImpl *>(pathseg)->x();
			cury += static_cast<SVGPathSegLinetoRelImpl *>(pathseg)->y();
			break;
		case PATHSEG_LINETO_HORIZONTAL_ABS:
			curx = static_cast<SVGPathSegLinetoHorizontalAbsImpl *>(pathseg)->x();
			break;
		case PATHSEG_LINETO_HORIZONTAL_REL:
			curx += static_cast<SVGPathSegLinetoHorizontalRelImpl *>(pathseg)->x();
			break;
		case PATHSEG_LINETO_VERTICAL_ABS:
			cury = static_cast<SVGPathSegLinetoVerticalAbsImpl *>(pathseg)->y();
			break;
		case PATHSEG_LINETO_VERTICAL_REL:
			cury += static_cast<SVGPathSegLinetoVerticalRelImpl *>(pathseg)->y();
			break;
		case PATHSEG_ARC_ABS:
			curx = static_cast<SVGPathSegArcAbsImpl *>(pathseg)->x();
			cury = static_cast<SVGPathSegArcAbsImpl *>(pathseg)->y();
			break;
		case PATHSEG_ARC_REL:
			curx += static_cast<SVGPathSegArcRelImpl *>(pathseg)->x();
			cury += static_cast<SVGPathSegArcRelImpl *>(pathseg)->y();
			break;
		case PATHSEG_CURVETO_CUBIC_ABS:
			curx = static_cast<SVGPathSegCurvetoCubicAbsImpl *>(pathseg)->x();
			cury = static_cast<SVGPathSegCurvetoCubicAbsImpl *>(pathseg)->y();
			break;
		case PATHSEG_CURVETO_CUBIC_REL:
			curx += static_cast<SVGPathSegCurvetoCubicRelImpl *>(pathseg)->x();
			cury += static_cast<SVGPathSegCurvetoCubicRelImpl *>(pathseg)->y();
			break;
		case PATHSEG_CURVETO_CUBIC_SMOOTH_ABS:
			curx = static_cast<SVGPathSegCurvetoCubicSmoothAbsImpl *>(pathseg)->x();
			cury = static_cast<SVGPathSegCurvetoCubicSmoothAbsImpl *>(pathseg)->y();
			break;
		case PATHSEG_CURVETO_CUBIC_SMOOTH_REL:
			curx += static_cast<SVGPathSegCurvetoCubicSmoothRelImpl *>(pathseg)->x();
			cury += static_cast<SVGPathSegCurvetoCubicSmoothRelImpl *>(pathseg)->y();
			break;
		case PATHSEG_CURVETO_QUADRATIC_ABS:
			curx = static_cast<SVGPathSegCurvetoQuadraticAbsImpl *>(pathseg)->x();
			cury = static_cast<SVGPathSegCurvetoQuadraticAbsImpl *>(pathseg)->y();
			break;
		case PATHSEG_CURVETO_QUADRATIC_REL:
			curx += static_cast<SVGPathSegCurvetoQuadraticRelImpl *>(pathseg)->x();
			cury += static_cast<SVGPathSegCurvetoQuadraticRelImpl *>(pathseg)->y();
			break;
		case PATHSEG_CURVETO_QUADRATIC_SMOOTH_ABS:
			curx = static_cast<SVGPathSegCurvetoQuadraticSmoothAbsImpl *>(pathseg)->x();
			cury = static_cast<SVGPathSegCurvetoQuadraticSmoothAbsImpl *>(pathseg)->y();
			break;
		case PATHSEG_CURVETO_QUADRATIC_SMOOTH_REL:
			curx += static_cast<SVGPathSegCurvetoQuadraticSmoothRelImpl *>(pathseg)->x();
			cury += static_cast<SVGPathSegCurvetoQuadraticSmoothRelImpl *>(pathseg)->y();
			break;
	}
}

// vim:ts=4:noet
